import logging

from .configmanager import ConfigManager
from .models import Config

logger = logging.getLogger(__name__)

def get_config():
    return ConfigManager.get_instance().config

def load_config(path) -> Config:
    return ConfigManager.get_instance(path).config



from packaging import version
from rich.console import Console

from .common import reboot_hosts
from .models import MikrotikHost

from mikrotools.netapi import MikrotikManager
from mikrotools.tools.colors import fcolors_256 as fcolors

def is_upgradable(current_version, upgrade_version):
    """
    Checks if the given `current_version` is upgradable to `upgrade_version`.

    Args:
        current_version (str): The current version of the host.
        upgrade_version (str): The version to check against.

    Returns:
        bool: True if the version is upgradable, False otherwise.
    """
    if current_version and upgrade_version:
        return version.parse(current_version) < version.parse(upgrade_version)

def print_check_upgradable_progress(host, counter, total, outdated, offline, failed=0):
    console = Console()
    
    if offline > 0:
        offline_color = 'red'
    else:
        offline_color = 'green'
    
    if failed > 0:
        failed_color = 'red'
    else:
        failed_color = 'green'
    
    print('\r\033[K', end='\r')
    console.print(f'[grey27]Checking host [sky_blue2]'
                  f'{host.identity if host.identity is not None else "-"} '
                  f'[cyan]([yellow]{host.address}[cyan]) '
                  f'[red]\\[{counter}/{total}] '
                  f'[medium_purple1]| [cyan]Upgradable: [medium_purple1]{outdated} '
                  f'[medium_purple1]| [cyan]Offline: [{offline_color}]{offline} '
                  f'[medium_purple1]| [cyan]Errors: [{failed_color}]{failed}',
                  end=''
                  )

def print_upgrade_progress(host, counter, total, remaining):
        print(f'\r{fcolors.darkgray}Upgrading {fcolors.lightblue}{host.identity} '
              f'{fcolors.blue}({fcolors.yellow}{host.address}{fcolors.blue}) '
              f'{fcolors.red}[{counter}/{total}] '
              f'{fcolors.cyan}Remaining: {fcolors.lightpurple}{remaining}{fcolors.default}'
              f'\033[K',
              end='')

def get_firmware_upgradable_hosts(addresses):
    upgradable_hosts = []
    failed = 0
    offline = 0
    counter = 1
    
    for address in addresses:
        host = MikrotikHost(address=address)
        print_check_upgradable_progress(host, counter, len(addresses), len(upgradable_hosts), offline, failed)
        
        try:
            with MikrotikManager.get_connection(address) as device:
                host.identity = device.get_identity()
                print_check_upgradable_progress(host, counter, len(addresses), len(upgradable_hosts), offline, failed)
                host.current_firmware_version = device.get_current_firmware_version()
                host.upgrade_firmware_version = device.get_upgrade_firmware_version()
        except TimeoutError:
            offline += 1
            counter += 1
            continue
        except Exception as e:
            failed += 1
            counter += 1
            continue
        
        if is_upgradable(host.current_firmware_version, host.upgrade_firmware_version):
            upgradable_hosts.append(host)
        
        counter += 1
    
    print('\r\033[K', end='\r')
    
    return upgradable_hosts

def get_routeros_upgradable_hosts(addresses) -> list[MikrotikHost]:
    upgradable_hosts = []
    failed = 0
    offline = 0
    counter = 1
    
    for address in addresses:
        host = MikrotikHost(address=address)
        print_check_upgradable_progress(host, counter, len(addresses), len(upgradable_hosts), offline, failed)
        try:
            with MikrotikManager.get_connection(address) as device:
                host.identity = device.get_identity()
                print_check_upgradable_progress(host, counter, len(addresses), len(upgradable_hosts), offline, failed)
                device.execute_command_raw('/system package update check-for-updates')
                host.installed_routeros_version = device.get_routeros_installed_version()
                host.latest_routeros_version = device.get('/system package update', 'latest-version')
        except TimeoutError:
            offline += 1
            counter += 1
            continue
        except Exception as e:
            failed += 1
            counter += 1
            continue
        
        if is_upgradable(host.installed_routeros_version, host.latest_routeros_version):
            upgradable_hosts.append(host)
        
        counter += 1
    
    print('\r\033[K', end='\r')
    
    return upgradable_hosts

def upgrade_hosts_firmware_start(addresses):
    """
    Starts the process of upgrading the firmware of the given hosts.

    This function checks which hosts have outdated firmware and prompts the user to
    confirm whether they want to upgrade them.

    Args:
        addresses (list[str]): A list of IP addresses or hostnames to check.
    """
    upgradable_hosts = get_firmware_upgradable_hosts(addresses)
    upgrade_hosts_firmware_confirmation_prompt(upgradable_hosts)

def upgrade_hosts_firmware_confirmation_prompt(upgradable_hosts):
    """
    Prompts the user to confirm whether they want to upgrade the specified hosts.

    Prints the list of hosts that will be upgraded and their respective
    identities. Then prompts the user to answer with 'y' to proceed with the
    upgrade or 'n' to exit the program.

    Args:
        upgradable_hosts (list[MikrotikHost]): A list of dictionaries each containing the
            information of a host to be upgraded.
    """
    # Checks if there are any hosts to upgrade
    if len(upgradable_hosts) == 0:
        print(f'{fcolors.bold}{fcolors.green}No hosts to upgrade firmware{fcolors.default}')
        exit()
    
    # Prints the list of hosts that will be upgraded
    print(f'{fcolors.bold}{fcolors.yellow}Upgradable hosts: {fcolors.red}{len(upgradable_hosts)}{fcolors.default}')
    print(f'\nThe following list of devices will be upgraded:\n')
    
    for host in upgradable_hosts:
        print(f'{fcolors.lightblue}Host: {fcolors.bold}{fcolors.green}{host.identity}'
              f'{fcolors.default} ({fcolors.lightpurple}{host.address}{fcolors.default})'
              f' {fcolors.blue}[{fcolors.red}{host.current_firmware_version} > {fcolors.green}{host.upgrade_firmware_version}{fcolors.blue}]'
              f'{fcolors.default}')

    # Prompts the user if they want to proceed
    print(f'\n{fcolors.bold}{fcolors.yellow}Are you sure you want to proceed? {fcolors.red}[y/N]{fcolors.default}')
    answer = input()
    
    # Continues or exits the program
    if answer.lower() == 'y':
        upgrade_hosts_firmware_apply(upgradable_hosts)
    else:
        exit()

def upgrade_hosts_firmware_apply(hosts):
    """
    Applies the firmware upgrade to all specified hosts and provides an option
    to reboot them afterward.

    For each host in the list, this function prints the upgrade progress, upgrades
    the host's firmware, and then offers the user the choice to reboot the devices.

    Args:
        hosts (list[MikrotikHost]): A list of MikrotikHost objects representing
            the hosts to be upgraded.

    Returns:
        None
    """

    counter = 1
    for host in hosts:
        print_upgrade_progress(host, counter, len(hosts), len(hosts) - counter + 1)
        upgrade_host_firmware(host)
        counter += 1
    
    print(f'\r\033[K', end='\r')
    print(f'{fcolors.bold}{fcolors.green}All hosts upgraded successfully!{fcolors.default}')
    print(f'{fcolors.bold}{fcolors.yellow}Would you like to reboot devices now? {fcolors.red}[y/n]{fcolors.default}')
    while True:
        answer = input()
        if answer.lower() == 'y':
            reboot_hosts(hosts)
            break
        elif answer.lower() == 'n':
            exit()
        else:
            print(f'{fcolors.bold}{fcolors.yellow}Invalid input. Please enter "y" or "n".{fcolors.default}')

def upgrade_host_firmware(host):
    """
    Upgrades the firmware of the specified host.

    :param host: A MikrotikHost object representing the host to upgrade.
    :return: None
    """
    try:
        with MikrotikManager.get_connection(host.address) as device:
            device.execute_command_raw('/system routerboard upgrade')
    except Exception as e:
        pass

def upgrade_hosts_routeros_start(addresses: list[str]) -> None:
    """
    Starts the process of upgrading RouterOS on the given hosts.

    This function checks which hosts have outdated RouterOS and prompts the user to
    confirm whether they want to upgrade them.

    Args:
        addresses (list[str]): A list of IP addresses or hostnames to check.
    """
    upgradable_hosts = get_routeros_upgradable_hosts(addresses)
    upgrade_hosts_routeros_confirmation_prompt(upgradable_hosts)

def upgrade_hosts_routeros_confirmation_prompt(upgradable_hosts: list[MikrotikHost]) -> None:
    # Checks if there are any hosts to upgrade
    """
    Prompts the user to confirm whether they want to upgrade the specified hosts.

    Prints the list of hosts that will be upgraded and their respective
    identities. Then prompts the user to answer with 'y' to proceed with the
    upgrade or 'n' to exit the program.

    Args:
        upgradable_hosts (list[MikrotikHost]): A list of objects each containing the
            information of a host to be upgraded.
    """
    if len(upgradable_hosts) == 0:
        print(f'{fcolors.bold}{fcolors.green}No hosts to upgrade RouterOS{fcolors.default}')
        exit()
    
    # Prints the list of hosts that will be upgraded
    print(f'{fcolors.bold}{fcolors.yellow}Upgradable hosts: {fcolors.red}{len(upgradable_hosts)}{fcolors.default}')
    print(f'\nThe following list of devices will be upgraded:\n')
    
    for host in upgradable_hosts:
        print(f'{fcolors.lightblue}Host: {fcolors.bold}{fcolors.green}{host.identity}'
              f'{fcolors.default} ({fcolors.lightpurple}{host.address}{fcolors.default}) '
              f'{fcolors.blue}[{fcolors.red}{host.installed_routeros_version} > {fcolors.green}{host.latest_routeros_version}{fcolors.blue}]'
              f'{fcolors.default}')

    print(f'\n{fcolors.bold}{fcolors.yellow}Are you sure you want to proceed? {fcolors.red}[y/N]{fcolors.default}')
    answer = input()
    
    if answer.lower() == 'y':
        upgrade_hosts_routeros_apply(upgradable_hosts)
    else:
        exit()

def upgrade_hosts_routeros_apply(hosts: list[MikrotikHost]) -> None:
    """
    Upgrades RouterOS version on all specified hosts.

    For each host in the list, this function prints the upgrade progress, upgrades
    the host's RouterOS version, and then prints a success message.

    Args:
        hosts (list[MikrotikHost]): A list of MikrotikHost objects representing
            the hosts to be upgraded.

    Returns:
        None
    """
    counter = 1
    for host in hosts:
        print_upgrade_progress(host, counter, len(hosts), len(hosts) - counter + 1)
        upgrade_host_routeros(host)
        counter += 1
    
    print(f'\r\033[K', end='\r')
    print(f'{fcolors.bold}{fcolors.green}All hosts upgraded successfully!{fcolors.default}')

def upgrade_host_routeros(host: MikrotikHost) -> None:
    """
    Upgrades the RouterOS version on the specified host.

    This function connects to the given host using SSH and initiates the RouterOS
    package update installation.

    Args:
        host (MikrotikHost): A MikrotikHost object representing the host to upgrade.
                             The host's address is used to establish the connection.

    Returns:
        None
    """
    try:
        with MikrotikManager.get_connection(host.address) as device:
            device.execute_command_raw('/system package update check-for-updates')
    except Exception as e:
        pass



from packaging import version
from rich.console import Console

from mikrotools.netapi import MikrotikManager
from mikrotools.tools.colors import fcolors_256 as fcolors

from .models import MikrotikHost

def get_device_config(host, sensitive=False):
    # Exporting current config
    with MikrotikManager.get_connection(host=host.address) as device:
        if sensitive:
            # Exporting sensitive config
            if version.parse(host.installed_routeros_version) >= version.parse('7.0'):
                # RouterOS 7.0+
                current_config = device.execute_command_raw('/export show-sensitive')
            else:
                # RouterOS < 7.0
                current_config = device.execute_command_raw('/export')
        else:
            # Exporting non-sensitive config
            if version.parse(host.installed_routeros_version) >= version.parse('7.0'):
                # RouterOS 7.0+
                current_config = device.execute_command_raw('/export')
            else:
                # RouterOS < 7.0
                current_config = device.execute_command_raw('/export hide-sensitive')
        
    return current_config

def backup_configs(addresses, sensitive=False):
    counter = 1
    failed_hosts = []
    
    for address in addresses:
        host = MikrotikHost(address=address)
        try:
            with MikrotikManager.get_connection(host=address) as device:
                host.identity = device.get_identity()
                host.installed_routeros_version = device.get_routeros_installed_version()
                
                print_backup_progress(host, counter, len(addresses), len(addresses) - counter + 1)
                
                current_config = get_device_config(host, sensitive)
        except Exception as e:
            failed_hosts.append(host)
            continue
        
        # Writing current config to file
        with open(f'{host.identity}.rsc', 'w') as f:
            f.write(current_config)
        
        counter += 1
    
    console = Console(highlight=False)
    print(f'\r\033[K', end='\r')

    if len(failed_hosts) > 0:
        console.print(f'[bold orange1]Backup completed with errors!\n'
                       f'[bold gold1]Backed up {len(addresses) - len(failed_hosts)} '
                       f'hosts out of {len(addresses)}\n')
        console.print(f'[bold red3]The following hosts failed to backup:')
        for host in failed_hosts:
            console.print(f'[thistle1]{host.address}')
    else:
        console.print(f'[bold green]All hosts backed up successfully!')

def print_backup_progress(host, counter, total, remaining):
    print(f'\r{fcolors.darkgray}Backing up {fcolors.lightblue}{host.identity} '
        f'{fcolors.blue}({fcolors.yellow}{host.address}{fcolors.blue}) '
        f'{fcolors.red}[{counter}/{total}]'
        f'{fcolors.cyan} Remaining: {fcolors.lightpurple}{remaining}{fcolors.default}'
        f'\033[K',
        end='')



from packaging import version

from mikrotools.netapi import MikrotikManager
from mikrotools.tools.colors import fcolors_256 as fcolors

def execute_hosts_commands(hosts, commands):
    for host in hosts:
        # Printing separator
        print(f'{fcolors.bold}{fcolors.lightblue}{"-"*30}{fcolors.default}')
        print(f'{fcolors.bold}{fcolors.lightblue}Working with host: {fcolors.lightpurple}{host}{fcolors.default}')
        
        with MikrotikManager.get_connection(host) as device:
            identity = device.get_identity()
            print(f'{fcolors.bold}{fcolors.lightblue}Identity: {fcolors.lightpurple}{identity}{fcolors.default}')
            installed_version = device.get_routeros_installed_version()
            print(f'{fcolors.bold}{fcolors.lightblue}Installed version: {fcolors.lightpurple}{installed_version}{fcolors.default}')
            
            # Executing commands
            for command in commands:
                print(f'\n{fcolors.bold}{fcolors.darkgray}Executing command: {command}{fcolors.default}')
                result = device.execute_command_raw(command)
                # Printing execution result
                print(result)

def get_outdated_hosts(hosts, min_version, filtered_version):

    """
    Checks the installed version of each host in the given list against the minimum
    version specified and returns a list of hosts with outdated versions.

    Args:
        hosts (list[str]): A list of hostnames or IP addresses to check.
        min_version (str): The minimum version required.
        filtered_version (str, optional): An optional version that further filters
                                          the hosts. If specified, the installed
                                          version must be greater than or equal
                                          to this version.

    Returns:
        list[str]: A list of hostnames or IP addresses with outdated versions.
    """
    counter = 1
    offline = 0
    outdated_hosts = []
    for host in hosts:
        print_progress(host, counter, len(hosts), len(outdated_hosts), offline)

        try:
            with MikrotikManager.get_connection(host) as device:
                installed_version = device.get_routeros_installed_version()
        except TimeoutError:
            offline += 1
            counter += 1
            continue
        
        if check_if_update_applicable(installed_version, min_version, filtered_version):
            outdated_hosts.append(host)
        
        counter += 1
    
    print('\r\033[K', end='\r')
    
    return outdated_hosts

def print_progress(host, counter, total, outdated, offline):
        print(f'\r{fcolors.darkgray}Checking host {fcolors.yellow}{host} '
            f'{fcolors.red}[{counter}/{total}] '
            f'{fcolors.cyan}Outdated: {fcolors.lightpurple}{outdated}{fcolors.default} '
            f'{fcolors.cyan}Offline: {fcolors.red}{offline}{fcolors.default}',
            end='')

def check_if_update_applicable(installed_version, min_version, filtered_version=None):
    """
    Checks the installed version of a host against the minimum version specified
    and returns True if an update is applicable, False otherwise.

    Args:
        host (str): The hostname or IP address of the host to check.
        min_version (str): The minimum version required.
        filtered_version (str, optional): An optional version that further filters
                                          the host. If specified, the installed
                                          version must be greater than or equal
                                          to this version.

    Returns:
        bool: True if an update is applicable, False otherwise.
    """
    
    installed_version = version.parse(installed_version)
    
    if installed_version < version.parse(min_version):
        if filtered_version:
            return installed_version >= version.parse(filtered_version)
        else:
            return True
    else:
        return False



#!/usr/bin/env python3

import click
import logging

from functools import wraps

from .config import get_config, load_config
from .tools.config import get_commands, get_hosts
from .tools.outputs import list_outdated_hosts
from .tools.ssh import execute_hosts_commands
from .tools.ssh import get_outdated_hosts
from .hoststools import backup_configs
from .hoststools.common import list_hosts, reboot_addresses
from .hoststools.upgrade import upgrade_hosts_firmware_start, upgrade_hosts_routeros_start
from .netapi import MikrotikManager

class Mutex(click.Option):
    def __init__(self, *args, **kwargs):
        self.not_required_if:list = kwargs.pop("not_required_if")

        assert self.not_required_if, "'not_required_if' parameter required"
        kwargs["help"] = (kwargs.get("help", "") + "Option is mutually exclusive with " + ", ".join(self.not_required_if) + ".").strip()
        super(Mutex, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        current_opt:bool = self.name in opts
        for mutex_opt in self.not_required_if:
            if mutex_opt in opts:
                if current_opt:
                    raise click.UsageError("Illegal usage: '" + str(self.name) + "' is mutually exclusive with " + str(mutex_opt) + ".")
                else:
                    self.required = False
        return super(Mutex, self).handle_parse_result(ctx, opts, args)

def common_options(func):
    @click.option('-H', '--host', help='Target host')
    @click.option('-P', '--port', type=int, help='SSH port')
    @click.option('-u', '--user', help='Username')
    @click.option('-p', '--password', is_flag=True, help='Prompt for password')
    @click.option('-c', '--config-file', default='settings.yaml')
    @click.option('-i', '--inventory-file')
    @click.option('-j', '--jump', is_flag=True, help='Use jump host')
    @click.option('-d', '--debug', is_flag=True, help='Enable debug mode')
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)
    return wrapper

def mikromanager_init(f):
    @wraps(f)
    def wrapper(port, user, password, config_file, inventory_file, jump, *args, **kwargs):
        logger = logging.getLogger(__name__)
        config = load_config(config_file)
    
        if port is not None:
            config.ssh.port = int(port)
        if user is not None:
            config.ssh.username = user
        if password:
            config.ssh.keyfile = None
            # Password prompt
            config.ssh.password = click.prompt('Password', hide_input=True)
        if inventory_file is not None:
            config.inventory.hostsFile = inventory_file
        if jump:
            config.ssh.jump = True
        
        logger.debug(f'Config after applying command line options: {config}')
        
        # Configuring MikrotikManager
        MikrotikManager.configure(config)
        
        return f(*args, **kwargs)
    
    return wrapper

def setup_logging(debug):
    if debug:
        level = logging.DEBUG
    else:
        level = logging.WARNING
    
    logging.basicConfig(
        level=level,
        format='%(asctime)s [%(levelname)s] %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S'
    )

def validate_commands(ctx, param, values):
    execute_command = ctx.params.get('execute_command')
    commands_file = ctx.params.get('commands_file')
    
    if (not execute_command and not commands_file):
        raise click.UsageError('You must provide either -e or -C')
    
    if (execute_command and commands_file):
        raise click.UsageError('You must provide either -e or -C, but not both.')
    
    return values

@click.group(invoke_without_command=True, context_settings=dict(help_option_names=['-h', '--help']))
@click.option('-e', '--execute-command', cls=Mutex, not_required_if=['commands_file'])
@click.option('-C', '--commands-file', cls=Mutex, not_required_if=['execute_command'])
@common_options
@click.pass_context
def cli(ctx, *args, **kwargs):
    # Setting up logging
    setup_logging(ctx.params['debug'])
    
    # Invoking default command
    if ctx.invoked_subcommand is None:
        validate_commands(ctx, None, None)
        ctx.invoke(execute, *args, **kwargs)

@cli.command(help='Backup configs from hosts')
@click.option('-s', '--sensitive', is_flag=True, default=False)
@mikromanager_init
@common_options
def backup(sensitive, *args, **kwargs):
    hosts = get_hosts()
    backup_configs(hosts, sensitive)

@cli.command(name='exec', help='Execute commands on hosts')
@click.option('-e', '--execute-command', cls=Mutex, not_required_if=['commands_file'])
@click.option('-C', '--commands-file', cls=Mutex, not_required_if=['execute_command'])
@mikromanager_init
@common_options
def execute(*args, **kwargs):
    hosts = get_hosts()
    
    # Getting command from arguments or config file
    commands = get_commands()
    
    # Executing commands for each host in list
    execute_hosts_commands(hosts, commands)

@cli.command(help='Check for routers with outdated firmware')
@click.argument('min-version')
@click.argument('filtered-version', required=False)
@click.option('-o', '--output-file', required=False)
@mikromanager_init
@common_options
def outdated(min_version, filtered_version, output_file, *args, **kwargs):
    hosts = get_hosts()
    outdated_hosts = get_outdated_hosts(hosts, min_version, filtered_version)
    if output_file:
        with open(output_file, 'w') as output_file:
            for host in outdated_hosts:
                output_file.write(f'{host}\n')
    else:
        list_outdated_hosts(outdated_hosts)

@cli.command(name='list', help='List routers')
@mikromanager_init
@common_options
def list_routers(*args, **kwargs):
    hosts = get_hosts()
    list_hosts(hosts)

@cli.command(help='Reboot routers')
@mikromanager_init
@common_options
def reboot(*args, **kwargs):
    addresses = get_hosts()
    reboot_addresses(addresses)

@cli.command(help='Upgrade routers with outdated RouterOS')
@mikromanager_init
@common_options
def upgrade(*args, **kwargs):
    hosts = get_hosts()
    upgrade_hosts_routeros_start(hosts)

@cli.command(help='Upgrade routers with outdated firmware')
@mikromanager_init
@common_options
def upgrade_firmware(*args, **kwargs):
    hosts = get_hosts()
    upgrade_hosts_firmware_start(hosts)

if __name__ == '__main__':
    cli()



from pydantic import BaseModel

class Base(BaseModel):
    pass

class Host(Base):
    address: str
    port: int | None = None
    username: str | None = None
    password: str | None = None
    keyfile: str | None = None



import threading

from contextlib import contextmanager
from functools import lru_cache
from paramiko.ssh_exception import SSHException
from typing import Generator

from mikrotools.config import Config

from .client import MikrotikSSHClient

class MikrotikManager:
    _config = None
    _connections = {}
    _lock = threading.Lock()
    
    @classmethod
    def configure(cls, config: Config):
        cls._config = config
        cls._connections.clear()
        cls.get_connection.cache_clear()
    
    @classmethod
    @lru_cache(maxsize=32)
    def get_connection(cls, host: str) -> MikrotikSSHClient:
        with cls._lock:
            if host in cls._connections:
                client = cls._connections[host]
                if client and client.is_connected():
                    return client
                else:
                    del cls._connections[host]
            
            if not cls._config:
                raise RuntimeError('MikrotikManager is not configured')
            
            username = cls._config.ssh.username
            port = cls._config.ssh.port
            password = cls._config.ssh.password
            if cls._config.ssh.keyfile:
                keyfile = cls._config.ssh.keyfile
            else:
                keyfile = None
            
            try:
                client = MikrotikSSHClient(
                    host=host,
                    port=port,
                    username=username,
                    password=password,
                    keyfile=keyfile
                )
                client.connect()
                cls._connections[host] = client
                return client
            except Exception as e:
                raise e
    
    @classmethod
    def close_all(cls) -> None:
        with cls._lock:
            for host, client in list(cls._connections.items()):
                try:
                    client.disconnect()
                except Exception as e:
                    pass
                del cls._connections[host]
            cls.get_connection.cache_clear()
    
    @classmethod
    @contextmanager
    def session(cls, host: str) -> Generator[MikrotikSSHClient, None, None]:
        client = cls.get_connection(host)
        if not client or not client.is_connected():
            raise ConnectionError(f'No active connection to {host}')
        
        try:
            yield client
        except SSHException as e:
            with cls._lock:
                if host in cls._connections:
                    del cls._connections[host]
                cls.get_connection.cache_clear()
            client.disconnect()
            raise
        finally:
            pass



import yaml

from pydantic import BaseModel

class Base (BaseModel):
    pass

class Inventory(Base):
    hostsFile: str = None

class JumpHost(Base):
    address: str = None
    port: int = 22
    username: str = None
    password: str = None
    keyfile: str = None

class SSHConfig(Base):
    port: int = 22
    username: str = None
    password: str = None
    keyfile: str = None
    jump: bool = False
    jumphost: JumpHost = JumpHost()

class Config(Base):
    ssh: SSHConfig = SSHConfig()
    inventory: Inventory = Inventory()

    @classmethod
    def from_yaml(cls, path: str) -> 'Config':
        with open(path, 'r') as f:
            data = yaml.safe_load(f)
            return cls(**data)



from packaging import version
from paramiko import AuthenticationException
from rich import print as rprint
from rich.box import SIMPLE
from rich.console import Console
from rich.prompt import Confirm
from rich.table import Table

from .models import MikrotikHost

from mikrotools.netapi import MikrotikManager
from mikrotools.tools.colors import fcolors_256 as fcolors

def list_hosts(addresses):
    offline_hosts = 0
    console = Console()
    table = Table(title="[green]List of hosts", show_header=True, header_style="bold grey78", box=SIMPLE)
    
    table.add_column("Host", justify="left")
    table.add_column("Address", justify="left")
    table.add_column('Public Address', justify="left")
    table.add_column("RouterOS", justify="left")
    table.add_column("Firmware", justify="left")
    table.add_column("Model", justify="left")
    table.add_column("CPU %", justify="right")
    table.add_column("Uptime", justify="left")
    
    console.clear()
    console.print(table)
    
    for address in addresses:
        failed = False
        error_message = None
        host = MikrotikHost(address=address)
        
        try:
            with MikrotikManager.get_connection(address) as device:
                host.identity = device.get_identity()
                host.public_address = device.get('/ip cloud', 'public-address')
                host.installed_routeros_version = device.get_routeros_installed_version()
                host.current_firmware_version = device.get('/system routerboard', 'current-firmware')
                host.model = device.get('/system routerboard', 'model')
                host.cpu_load = int(device.get('/system resource', 'cpu-load'))
                if version.parse(host.installed_routeros_version) >= version.parse('7.0'):
                    host.uptime = device.get('/system resource', 'uptime as-string')
                else:
                    host.uptime = device.get('/system resource', 'uptime')
        except TimeoutError:
            failed = True
            error_message = 'Connection timeout'
            offline_hosts += 1
        except AuthenticationException:
            failed = True
            error_message = 'Authentication failed'
        except Exception as e:
            failed = True
            error_message = str(e)
        
        if failed:
            table.add_row(
                f'[red]{host.identity if host.identity is not None else "-"}', # Host
                f'[light_steel_blue1]{host.address if host.address is not None else "-"}', # Address
                f'[red]{error_message if error_message is not None else "-"}'
            )
        else:
            if host.cpu_load is None:
                cpu_color = 'red'
            elif host.cpu_load < 40:
                cpu_color = 'green'
            elif host.cpu_load < 60:
                cpu_color = 'yellow'
            elif host.cpu_load < 80:
                cpu_color = 'dark_orange'
            else:
                cpu_color = 'red'
            
            table.add_row(
                f'[dark_orange]{host.identity if host.identity is not None else "-"}', # Host
                f'[light_steel_blue1]{host.address if host.address is not None else "-"}', # Address
                f'[slate_blue1]{host.public_address if host.public_address is not None else "-"}', # Public address
                f'[dark_olive_green3]{host.installed_routeros_version if host.installed_routeros_version is not None else "-"}', # RouterOS
                f'[medium_purple1]{host.current_firmware_version if host.current_firmware_version is not None else "-"}', # Firmware
                f'[dodger_blue2]{host.model if host.model is not None else "-"}', # Model
                f'[{cpu_color}]{host.cpu_load if host.cpu_load is not None else "-"}%', # CPU %
                f'[cornflower_blue]{host.uptime if host.uptime is not None else "-"}', # Uptime
            )
        
        console.clear()
        console.print(table)
        
        console.print(f'[medium_purple1]{"-" * 15}')
        console.print(f'[cornflower_blue]Total hosts: '
                      f'[light_steel_blue1]{len(table.rows)} '
                      f'[medium_purple1]| '
                      f'[cornflower_blue]Online hosts: '
                      f'[green]{len(table.rows) - offline_hosts} '
                      f'[medium_purple1]| '
                      f'[cornflower_blue]Offline hosts: '
                      f'[red]{offline_hosts} '
                      f'\n')

def print_reboot_progress(host, counter, total, remaining):
    # Clears the current line
    print('\r\033[K', end='')
    # Prints the reboot progress
    rprint(f'[grey27]Rebooting [sky_blue2]{host.identity if host.identity is not None else "-"} '
            f'[blue]([yellow]{host.address}[blue]) '
            f'[red]\\[{counter}/{total}] '
            f'[cyan]Remaining: [medium_purple1]{remaining}',
            end=''
    )

def reboot_addresses(addresses):
    hosts = []

    print(f'The following hosts will be rebooted:')
    for address in addresses:
        rprint(f'[light_slate_blue]Host: [bold sky_blue1]{address}')
    
    is_reboot_confirmed = Confirm.ask(f'[bold yellow]Would you like to reboot devices now?[/] '
                                      f'[bold red]\\[y/n][/]', show_choices=False)

    if not is_reboot_confirmed:
        exit()
    else:
        [hosts.append(MikrotikHost(address=address)) for address in addresses]
        reboot_hosts(hosts)

def reboot_hosts(hosts):
    failed = 0
    failed_hosts = []
    counter = 1
    console = Console(highlight=False)
    
    for host in hosts:
        print_reboot_progress(host, counter, len(hosts), len(hosts) - counter + 1)
        try:
            reboot_host(host)
        except Exception as e:
            failed += 1
            failed_hosts.append(host)
        counter += 1
    
    print(f'\r\033[K', end='\r')
    print('')
    if failed > 0:
        console.print(f'[bold orange1]Rebooted {len(hosts) - failed} hosts out of {len(hosts)}!\n')
        console.print(f'[bold red3]The following hosts failed to reboot:')
        for host in failed_hosts:
            console.print(f'[grey78]{host.address}')
        exit()
    rprint(f'[bold green]All hosts rebooted successfully!')

def reboot_host(host):
    with MikrotikManager.get_connection(host.address) as device:
        device.execute_command('/system reboot')



from .base import Base, Host

class MikrotikHost(Host):
    identity: str | None = None
    installed_routeros_version: str | None = None
    latest_routeros_version: str | None = None
    current_firmware_version: str | None = None
    upgrade_firmware_version: str | None = None
    cpu_load: int | None = None
    model: str | None = None
    uptime: str | None = None
    public_address: str | None = None



from .models import Config

import logging

logger = logging.getLogger(__name__)

class ConfigManager:
    _instance = None
    
    def __init__(self, path: str):
        self.config = self.load_config(path)
    
    @classmethod
    def get_instance(cls, path: str = None):
        if cls._instance is None:
            cls._instance = cls(path)
        return cls._instance
    
    def load_config(self, path: str) -> Config:
        try:
            config = Config.from_yaml(path)
            logger.debug(f'Config loaded from YAML: {path}')
        except FileNotFoundError:
            config = Config()
            logger.warning(f'Config file not found: {path}, using default config')
        except TypeError:
            config = Config()
            logger.warning(f'Config file is empty or invalid: {path}, using default config')
        
        logger.debug(f'Config loaded: {config}')
        return config



import paramiko

class MikrotikSSHClient():
    def __init__(self, host: str, username: str, password: str = None, keyfile: str = None, port: int = 22):
        self._host = host
        self._port = port
        self._username = username
        self._password = password
        self._keyfile = keyfile
        self._ssh = paramiko.SSHClient()
        self._ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self._connected = False
    
    def connect(self) -> None:
        disabled_algorithms = {'pubkeys': ['rsa-sha2-256', 'rsa-sha2-512']}
        timeout = 5
        
        # Check if password or keyfile is provided
        if self._password is not None and self._keyfile == None:
            # Connect with password
            try:
                self._ssh.connect(
                    self._host,
                    port=self._port,
                    username=self._username,
                    password=self._password,
                    disabled_algorithms=disabled_algorithms,
                    timeout=timeout,
                    look_for_keys=False
                )
                self._connected = True
            except Exception as e:
                raise e
        elif self._keyfile is not None and self._password == None:
            # Connect with key
            try:
                self._ssh.connect(
                    self._host,
                    port=self._port,
                    username=self._username,
                    key_filename=self._keyfile,
                    disabled_algorithms=disabled_algorithms,
                    timeout=timeout,
                    look_for_keys=False
                )
                self._connected = True
            except Exception as e:
                raise e
        else:
            raise Exception('Must provide either password or keyfile')
    
    def disconnect(self) -> None:
        if self._connected:
            self._ssh.close()
            self._connected = False
    
    def execute_command(self, command: str) -> list[str]:
        """
        Execute a command on the Mikrotik device and return its output as a list of strings.

        :param command: The command to execute
        :return: The output of the command
        :raises: ConnectionError if not connected to the host
                 RuntimeError if the command execution fails
        """
        output = self.execute_command_raw(command).strip().split('\n')
        
        return [line.strip() for line in output if line.strip()]
    
    def execute_command_raw(self, command: str) -> str:
        """
        Execute a command on the Mikrotik device and return its output as a raw string.

        :param command: The command to execute
        :return: The output of the command
        :raises: ConnectionError if not connected to the host
                 RuntimeError if the command execution fails
        """
        if not self._connected:
            raise ConnectionError('Not connected to host')
        
        try:
            _, stdout, stderr = self._ssh.exec_command(command)
            output = stdout.read().decode()
            error = stderr.read().decode().strip()
            
            if error:
                raise RuntimeError(f'Error executing command: {error}')
            
            return output
        except paramiko.SSHException as e:
            raise e
    
    def get(self, path: str, obj: str = None) -> str:
        """
        Retrieves an object from a path on the router.

        Args:
            path: The path to the object
            obj: The object to retrieve

        Returns:
            The value of the object as a string
        """
        # Remove trailing slash
        path = path.rstrip('/')
        
        if obj:
            path = f'{path} get {obj}'
        else:
            path = f'{path} get'
        return self.execute_command_raw(f':put [{path}]').strip()
    
    def get_dict(self, path: str, obj: str = None) -> list[str]:
        """
        Retrieves a dictionary representation of an object's properties from a path on the router.

        Args:
            path: The path to the object.
            obj: The object to retrieve. Optional, if not specified, retrieves default object.

        Returns:
            A dictionary where keys are the object's property names and values are the corresponding
            property values, extracted from the response string.
        """
        output = {}
        current_key = None
        
        response = self.get(path, obj)
        
        for part in response.split(';'):
            part = part.strip()
            if not part:
                continue
            
            if '=' in part:
                key, value = part.split('=', 1)
                current_key = key.strip()
                output[current_key] = value.strip()
            elif current_key and current_key in output:
                output[current_key] += ';' + part.strip()
        
        return output
        
    def get_identity(self) -> str:
        """
        Retrieves the identity of the router.

        Returns:
            The identity of the router as a string.
        """
        return self.get('/system identity', 'name')
    
    def get_routeros_installed_version(self) -> str:
        """
        Retrieves the currently installed version of RouterOS.

        Returns:
            The installed version of RouterOS as a string.
        """
        return self.get('/system package update', 'installed-version')

    def get_current_firmware_version(self) -> str:
        """
        Retrieves the current firmware version of the router.

        Returns:
            The current firmware version as a string.
        """
        return self.get('/system routerboard', 'current-firmware')
    
    def get_upgrade_firmware_version(self) -> str:
        """
        Retrieves the upgrade firmware version of the router.

        Returns:
            The upgrade firmware version as a string.
        """
        return self.get('/system routerboard', 'upgrade-firmware')
    
    def __enter__(self):
        self.connect()
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()



import click
import logging

from mikrotools.config import get_config

logger = logging.getLogger(__name__)

def get_commands():
    ctx = click.get_current_context()

    if ctx.params['execute_command']:
        commands = [ctx.params['execute_command']]
    elif ctx.params['commands_file']:
        commands = get_commands_from_file(ctx.params['commands_file'])
    else:
        commands = []
    
    return commands

def get_commands_from_file(filename):
    with open(filename) as commands_file:
        commands = [command.rstrip() for command in commands_file]
        return commands

def get_hosts():
    ctx = click.get_current_context()
    if ctx.params['host']:
        hosts = [ctx.params['host']]
    elif ctx.params['inventory_file']:
        logger.debug(f'get_hosts: Inventory file path set from command line: '
                     f'{ctx.params["inventory_file"]}')
        try:
            hosts = read_hosts_from_file(ctx.params['inventory_file'])
        except TypeError:
            logger.error('Inventory file or host address is not specified')
            exit(1)
        except FileNotFoundError:
            logger.error(f'Inventory file not found: {ctx.params["inventory_file"]}')
            exit(1)
    else:
        # Getting config from YAML file
        config = get_config()
        logger.debug(f'get_hosts: Config: {config}')
        logger.debug(f'get_hosts: Inventory file path set from config: '
                     f'{config.inventory.hostsFile}')
        try:
            hosts = read_hosts_from_file(config.inventory.hostsFile)
        except TypeError:
            logger.error('Inventory file or host address is not specified')
            exit(1)
        except FileNotFoundError:
            logger.error(f'Inventory file not found: {config.inventory.hostsFile}')
            exit(1)
    
    return hosts

def read_hosts_from_file(filename):
    with open(filename) as hostsfile:
        hosts = [host.rstrip() for host in hostsfile]
        return hosts



import argparse

def parse_args():
    parser = argparse.ArgumentParser(description='Simple NAS/GW managing script')
    parser.add_argument('-e', '--execute-command')
    parser.add_argument('-c', '--config-file')
    parser.add_argument('-i', '--inventory-file')
    parser.add_argument('-H', '--host')
    parser.add_argument('-C', '--commands-file')
    args = parser.parse_args()
    return args
